# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy

from numpy import zeros, ndarray, dot, sqrt, outer, insert, delete

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import get_stiffness_from_young_poisson
from pyfem.utils.colors import error_style


class PlasticKinematicHardening(BaseMaterial):
    """
    随动强化塑性材料。

    支持的截面属性：('Volume', 'PlaneStress', 'PlaneStrain')

    :ivar E: Young's modulus E
    :vartype E: float

    :ivar nu: Poisson's ratio nu
    :vartype nu: float

    :ivar yield_stress: Yield stress
    :vartype yield_stress: float

    :ivar hard: Hardening coefficient
    :vartype hard: float

    :ivar EBULK3: 3倍体积模量
    :vartype EBULK3: float

    :ivar EG: 剪切模量
    :vartype EG: float

    :ivar EG2: 2倍剪切模量
    :vartype EG2: float

    :ivar EG3: 3倍剪切模量
    :vartype EG3: float

    :ivar ELAM: 拉梅常数
    :vartype ELAM: float

    :ivar tolerance: 判断屈服的误差容限
    :vartype tolerance: float
    """

    __slots_dict__: dict = {
        'E': ('float', 'Young\'s modulus E'),
        'nu': ('float', 'Poisson\'s ratio nu'),
        'yield_stress': ('float', 'Yield stress'),
        'hard': ('float', 'Hardening coefficient'),
        'EBULK3': ('float', '3倍体积模量'),
        'EG': ('float', '剪切模量'),
        'EG2': ('float', '2倍剪切模量'),
        'EG3': ('float', '3倍剪切模量'),
        'ELAM': ('float', '拉梅常数'),
        'tolerance': ('float', '判断屈服的误差容限'),
    }

    __slots__ = BaseMaterial.__slots__ + [slot for slot in __slots_dict__.keys()]

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = ['Young\'s modulus E', 'Poisson\'s ratio nu', 'Yield stress', 'Hardening coefficient']

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.E: float = self.data_dict['Young\'s modulus E']
        self.nu: float = self.data_dict['Poisson\'s ratio nu']
        self.yield_stress: float = self.data_dict['Yield stress']
        self.hard: float = self.data_dict['Hardening coefficient']

        self.EBULK3: float = self.E / (1.0 - 2.0 * self.nu)
        self.EG2: float = self.E / (1.0 + self.nu)
        self.EG: float = self.EG2 / 2.0
        self.EG3: float = 3.0 * self.EG
        self.ELAM: float = (self.EBULK3 - self.EG2) / 3.0
        self.tolerance: float = 1.0e-10

        self.create_tangent()

    def create_tangent(self):
        if self.section.type in self.allowed_section_types:
            self.tangent = get_stiffness_from_young_poisson(self.dimension, self.E, self.nu, self.section.type)
        else:
            raise NotImplementedError(error_style(self.get_section_type_error_msg()))

    def get_tangent(self, variable: dict[str, ndarray],
                    state_variable: dict[str, ndarray],
                    state_variable_new: dict[str, ndarray],
                    element_id: int,
                    iqp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> tuple[ndarray, dict[str, ndarray]]:

        if state_variable == {} or timer.time0 == 0.0:
            state_variable['elastic_strain'] = zeros(ntens, dtype=DTYPE)
            state_variable['plastic_strain'] = zeros(ntens, dtype=DTYPE)
            state_variable['back_stress'] = zeros(ntens, dtype=DTYPE)
            state_variable['stress'] = zeros(ntens, dtype=DTYPE)

        elastic_strain = deepcopy(state_variable['elastic_strain'])
        plastic_strain = deepcopy(state_variable['plastic_strain'])
        back_stress = deepcopy(state_variable['back_stress'])
        stress = deepcopy(state_variable['stress'])

        dstrain = variable['dstrain']

        E = self.E
        nu = self.nu

        if self.section.type == 'PlaneStrain':
            dstrain = insert(dstrain, 2, 0)
        elif self.section.type == 'PlaneStress':
            dstrain = insert(dstrain, 2, -nu / (1 - nu) * (dstrain[0] + dstrain[1]))

        elastic_strain += dstrain

        ddsdde = zeros((ntens, ntens), dtype=DTYPE)

        lam = E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu))
        mu = E / (2.0 * (1.0 + nu))

        ddsdde[:ndi, :ndi] += lam
        for i in range(ndi):
            ddsdde[i, i] += 2 * mu
        for i in range(ndi, ntens):
            ddsdde[i, i] += mu

        stress += dot(ddsdde, dstrain)
        s = stress - back_stress
        smises = get_smises(s)

        # if element_id == 0 and iqp == 0:
        #     print(ddsdde)
        #     print(dstrain)
        #     print(stress)
        #     print(dot(ddsdde, dstrain))

        if smises > (1.0 + self.tolerance) * self.yield_stress:
            hydrostatic_stress = sum(stress[:ndi]) / 3.0
            flow = stress - back_stress
            flow[:ndi] = flow[:ndi] - hydrostatic_stress
            flow *= 1.0 / smises

            delta_p = (smises - self.yield_stress) / (self.EG3 + self.hard)
            back_stress += self.hard * flow * delta_p

            plastic_strain[:ndi] += 1.5 * flow[:ndi] * delta_p
            elastic_strain[:ndi] -= 1.5 * flow[:ndi] * delta_p

            plastic_strain[ndi:] += 3.0 * flow[ndi:] * delta_p
            elastic_strain[ndi:] -= 3.0 * flow[ndi:] * delta_p

            stress = back_stress + flow * self.yield_stress
            stress[:ndi] += hydrostatic_stress

            # if element_id == 0 and iqp == 0:
            #     print(stress)

            EFFG = self.EG * (self.yield_stress + self.hard * delta_p) / smises
            EFFG2 = 2.0 * EFFG
            EFFG3 = 3.0 * EFFG
            EFFLAM = (self.EBULK3 - EFFG2) / 3.0
            EFFHRD = self.EG3 * self.hard / (self.EG3 + self.hard) - EFFG3

            ddsdde = zeros(shape=(ntens, ntens), dtype=DTYPE)
            ddsdde[:ndi, :ndi] = EFFLAM

            for i in range(ndi):
                ddsdde[i, i] += EFFG2

            for i in range(ndi, ntens):
                ddsdde[i, i] += EFFG

            ddsdde += EFFHRD * outer(flow, flow)

        # if element_id == 0 and iqp == 0:
        #     print(stress)

        state_variable_new['elastic_strain'] = elastic_strain
        state_variable_new['plastic_strain'] = plastic_strain
        state_variable_new['back_stress'] = back_stress
        state_variable_new['stress'] = stress

        plastic_energy = sum(plastic_strain * stress)

        if self.section.type == 'PlaneStrain':
            ddsdde = delete(delete(ddsdde, 2, axis=0), 2, axis=1)
            stress = delete(stress, 2)
        elif self.section.type == 'PlaneStress':
            ddsdde = delete(delete(ddsdde, 2, axis=0), 2, axis=1)
            ddsdde[0, 0] -= lam * lam / (lam + 2 * mu)
            ddsdde[0, 1] -= lam * lam / (lam + 2 * mu)
            ddsdde[1, 0] -= lam * lam / (lam + 2 * mu)
            ddsdde[1, 1] -= lam * lam / (lam + 2 * mu)
            stress = delete(stress, 2)

        output = {'stress': stress, 'plastic_energy': plastic_energy}

        return ddsdde, output


def get_smises(s: ndarray) -> float:
    if len(s) == 3:
        smises = sqrt(s[0] ** 2 + s[1] ** 2 - s[0] * s[1] + 3 * s[2] ** 2)
        return float(smises)
    elif len(s) >= 3:
        smises = (s[0] - s[1]) ** 2 + (s[1] - s[2]) ** 2 + (s[2] - s[0]) ** 2
        smises += 6 * sum([i ** 2 for i in s[3:]])
        smises = sqrt(0.5 * smises)
        return float(smises)
    else:
        raise NotImplementedError(error_style(f'unsupported stress dimension {len(s)}'))


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(PlasticKinematicHardening.__slots_dict__)

    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\mechanical\plane\Job-1.toml')
    material_data = PlasticKinematicHardening(props.materials[0], 3, props.sections[0])
    material_data.show()



# -*- coding: utf-8 -*-
"""

"""
from numpy import meshgrid, outer, column_stack
from numpy.polynomial.legendre import leggauss

from pyfem.fem.constants import DTYPE
from pyfem.quadrature.BaseQuadrature import BaseQuadrature
from pyfem.utils.colors import error_style


class GaussLegendreQuadrature(BaseQuadrature):
    """
    基于 Gauss-Legendre 多项式计算笛卡尔坐标系下标准线段、正方形和立方体的积分点坐标和权重。

    标准线段::

       (-1)    (0)     (1)
        0---------------1
                +-->ξ

    标准正方形::

        (-1,1)           (1,1)
          3---------------2
          |       η       |
          |       |       |
          |       o--ξ    |
          |               |
          |               |
          0---------------1
        (-1,-1)          (1,-1)

    标准正六面体::

                     (-1,1,1)        (1,1,1)
                      7---------------6
                     /|              /|
                    / |     ζ  η    / |
                   /  |     | /    /  |
        (-1,-1,1) 4---+-----|/----5 (1,-1,1)
                  |   |     o--ξ  |   |
                  |   3-----------+---2 (1,1,-1)
                  |  /(-1,1,-1)   |  /
                  | /             | /
                  |/              |/
                  0---------------1
                 (-1,-1,-1)      (1,-1,-1)

    """

    __slots__ = BaseQuadrature.__slots__ + []

    def __init__(self, order: int, dimension: int) -> None:
        super().__init__(order, dimension)
        xi, weight = leggauss(order)
        if dimension == 1:
            xi = xi.reshape(len(xi), -1)
            weight = weight.reshape(len(weight), -1)

        elif dimension == 2:
            xi1, xi2 = meshgrid(xi, xi)
            xi1 = xi1.ravel()
            xi2 = xi2.ravel()
            xi = column_stack((xi1, xi2))
            weight = outer(weight, weight)
            weight = weight.ravel()

        elif dimension == 3:
            xi1, xi2, xi3 = meshgrid(xi, xi, xi)
            xi1 = xi1.ravel()
            xi2 = xi2.ravel()
            xi3 = xi3.ravel()
            xi = column_stack((xi1, xi2, xi3))
            weight = outer(outer(weight, weight), weight)
            weight = weight.ravel()

        else:
            raise NotImplementedError(error_style('dimension must be 1, 2 or 3'))

        self.qp_coords = xi.astype(DTYPE)
        self.qp_weights = weight.astype(DTYPE)
        self.qp_number = len(self.qp_weights)


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(GaussLegendreQuadrature.__slots_dict__)

    quadrature = GaussLegendreQuadrature(2, 2)
    quadrature.show()



# -*- coding: utf-8 -*-
"""

"""
from typing import Callable

from numpy import empty, array, ndarray, insert, in1d, sqrt

from pyfem.isoelements.IsoElementDiagram import IsoElementDiagram
from pyfem.isoelements.shape_functions import get_shape_line2, get_shape_tetra4, get_shape_empty, get_shape_hex20, \
    get_shape_quad4, get_shape_tria3, get_shape_line3, get_shape_quad8, get_shape_tria6, get_shape_hex8
from pyfem.quadrature.GaussLegendreQuadrature import GaussLegendreQuadrature
from pyfem.quadrature.TetrahedronQuadrature import TetrahedronQuadrature
from pyfem.quadrature.TriangleQuadrature import TriangleQuadrature
from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_slots_to_string_ndarray


class IsoElementShape:
    """
    等参单元类，设置等参单元的形函数和积分点等信息。

    当前支持的单元类型 ['empty', 'line2', 'line3', 'tria3', 'tria6', 'quad4', 'quad8', 'tetra4', 'hex8', 'hex20']

    :ivar element_type: 等参单元类型
    :vartype element_type: str

    :ivar diagram: 等参单元示意图（字符串形式）
    :vartype diagram: str

    :ivar dimension: 等参单元空间维度
    :vartype dimension: int

    :ivar topological_dimension: 等参单元拓扑维度
    :vartype topological_dimension: int

    :ivar nodes_number: 等参单元节点数目
    :vartype nodes_number: int

    :ivar order: 等参单元插值阶次
    :vartype order: int

    :ivar shape_function: 等参单元形函数
    :vartype shape_function: Callable

    :ivar qp_number: 等参单元积分点数量
    :vartype qp_number: int

    :ivar qp_coords: 等参单元积分点坐标
    :vartype qp_coords: ndarray

    :ivar qp_weights: 等参单元积分点权重
    :vartype qp_weights: ndarray

    :ivar qp_shape_values: 等参单元积分点处形函数的值
    :vartype qp_shape_values: ndarray

    :ivar qp_shape_gradients: 等参单元积分点处形函数对局部坐标梯度的值
    :vartype qp_shape_gradients: ndarray

    :ivar bc_surface_number: 等参单元边表面数量
    :vartype bc_surface_number: int

    :ivar bc_surface_nodes_dict: 等参单元边表面节点编号
    :vartype bc_surface_nodes_dict: dict[str, tuple]

    :ivar bc_surface_coord_dict: 等参单元边表面节点坐标
    :vartype bc_surface_coord_dict: dict[str, tuple]

    :ivar bc_qp_coords_dict: 等参单元边表面积分点坐标
    :vartype bc_qp_coords_dict: dict[str, ndarray]

    :ivar bc_qp_weights: 等参单元边表面积分点权重
    :vartype bc_qp_weights: ndarray

    :ivar bc_qp_shape_values_dict: 等参单元边表面积分点处形函数的值
    :vartype bc_qp_shape_values_dict: dict[str, ndarray]

    :ivar bc_qp_shape_gradients_dict: 等参单元边表面积分点处形函数对局部坐标梯度的值
    :vartype bc_qp_shape_gradients_dict: dict[str, ndarray]

    :ivar nodes_on_surface_dict: 单元节点与等参单元边表面的映射字典
    :vartype nodes_on_surface_dict: dict[str, ndarray]
    """

    __slots_dict__: dict = {
        'element_type': ('str', '等参单元类型'),
        'diagram': ('str', '等参单元示意图（字符串形式）'),
        'dimension': ('int', '等参单元空间维度'),
        'topological_dimension': ('int',
                                  '等参单元拓扑维度，有些情况下拓扑维度不等于空间维度，例如处理空间曲面单元时，空间维度为3，但是单元拓扑维度为2'),
        'nodes_number': ('int', '等参单元节点数目'),
        'order': ('int', '等参单元插值阶次'),
        'shape_function': ('Callable', '等参单元形函数'),
        'qp_number': ('int', '等参单元积分点数量'),
        'qp_coords': ('ndarray', '等参单元积分点坐标'),
        'qp_weights': ('ndarray', '等参单元积分点权重'),
        'qp_shape_values': ('ndarray', '等参单元积分点处形函数的值'),
        'qp_shape_gradients': ('ndarray', '等参单元积分点处形函数对局部坐标梯度的值'),
        'bc_surface_number': ('int', '等参单元边表面数量'),
        'bc_surface_nodes_dict': ('dict[str, tuple]', '等参单元边表面节点编号'),
        'bc_surface_coord_dict': ('dict[str, tuple]', '等参单元边表面节点坐标'),
        'bc_qp_coords_dict': ('dict[str, ndarray]', '等参单元边表面积分点坐标'),
        'bc_qp_weights': ('ndarray', '等参单元边表面积分点权重'),
        'bc_qp_shape_values_dict': ('dict[str, ndarray]', '等参单元边表面积分点处形函数的值'),
        'bc_qp_shape_gradients_dict': ('dict[str, ndarray]', '等参单元边表面积分点处形函数对局部坐标梯度的值'),
        'nodes_on_surface_dict': ('dict[str, ndarray]', '单元节点与等参单元边表面的映射字典')
    }

    __slots__: list = [slot for slot in __slots_dict__.keys()]

    allowed_element_type = ['empty', 'line2', 'line3', 'tria3', 'tria6', 'quad4', 'quad8', 'tetra4', 'hex8', 'hex20']

    def __init__(self, element_type: str) -> None:
        self.element_type: str = ''
        self.diagram: str = ''
        self.dimension: int = 0
        self.topological_dimension: int = 0
        self.nodes_number: int = 0
        self.order: int = 0
        self.shape_function: Callable = get_shape_empty
        self.qp_number: int = 0
        self.qp_coords: ndarray = empty(0)
        self.qp_weights: ndarray = empty(0)
        self.qp_shape_values: ndarray = empty(0)
        self.qp_shape_gradients: ndarray = empty(0)
        self.bc_surface_number: int = 0
        self.bc_surface_nodes_dict: dict[str, tuple] = dict()
        self.bc_surface_coord_dict: dict[str, tuple] = dict()
        self.bc_qp_coords_dict: dict[str, ndarray] = dict()
        self.bc_qp_weights: ndarray = empty(0)
        self.bc_qp_shape_values_dict: dict[str, ndarray] = dict()
        self.bc_qp_shape_gradients_dict: dict[str, ndarray] = dict()
        self.nodes_on_surface_dict: dict[str, ndarray] = dict()

        element_type_dict = {
            'line2': self.set_line2,
            'line3': self.set_line3,
            'tria3': self.set_tria3,
            'tria6': self.set_tria6,
            'quad4': self.set_quad4,
            'quad8': self.set_quad8,
            'tetra4': self.set_tetra4,
            'hex8': self.set_hex8,
            'hex20': self.set_hex20
        }

        self.element_type = element_type

        if element_type == 'empty':
            pass
        elif element_type in element_type_dict.keys():
            element_type_dict[element_type]()
        else:
            error_msg = f'unsupported element type {element_type}'
            raise NotImplementedError(error_style(error_msg))

        # 根据权重数组计算积分点数量
        self.qp_number = len(self.qp_weights)

        # 根据等参单元形函数，计算积分点处形函数的值和形函数梯度的值
        qp_shape_values = list()
        qp_shape_gradients = list()
        for qp_coord in self.qp_coords:
            N, dNdxi = self.shape_function(qp_coord)
            qp_shape_values.append(N)
            qp_shape_gradients.append(dNdxi)
        self.qp_shape_values = array(qp_shape_values)
        self.qp_shape_gradients = array(qp_shape_gradients)

        # 建立等参单元表面名称和单元节点是否在当前表面的映射关系
        for surface_name, surface_conn in self.bc_surface_nodes_dict.items():
            self.nodes_on_surface_dict[surface_name] = array(in1d(range(self.nodes_number), surface_conn))

        # 计算等参单元表面积分点处形函数的值和形函数梯度的值
        self.bc_qp_shape_values_dict = dict()
        self.bc_qp_shape_gradients_dict = dict()
        for bc_surface_name, bc_surface_qp_coords in self.bc_qp_coords_dict.items():
            bc_qp_shape_values = list()
            bc_qp_shape_gradients = list()
            for bc_surface_qp_coord in bc_surface_qp_coords:
                N, dNdxi = self.shape_function(bc_surface_qp_coord)
                bc_qp_shape_values.append(N)
                bc_qp_shape_gradients.append(dNdxi)
            self.bc_qp_shape_values_dict[bc_surface_name] = array(bc_qp_shape_values)
            self.bc_qp_shape_gradients_dict[bc_surface_name] = array(bc_qp_shape_gradients)

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def set_line2(self) -> None:
        self.dimension = 1
        self.topological_dimension = 1
        self.nodes_number = 2
        self.order = 1
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_line2
        self.diagram = IsoElementDiagram.line2

    def set_line3(self) -> None:
        self.dimension = 1
        self.topological_dimension = 1
        self.nodes_number = 3
        self.order = 2
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_line3
        self.diagram = IsoElementDiagram.line3

    def set_quad4(self) -> None:
        self.dimension = 2
        self.topological_dimension = 2
        self.nodes_number = 4
        self.order = 2
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_quad4

        self.bc_surface_number = 4
        bc_quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension - 1)
        bc_qp_coords, self.bc_qp_weights = bc_quadrature.get_quadrature_coords_and_weights()
        self.bc_surface_nodes_dict = {'s1': (3, 0),
                                      's2': (1, 2),
                                      's3': (0, 1),
                                      's4': (2, 3)}
        self.bc_surface_coord_dict = {'s1': (0, -1, -1, 1),
                                      's2': (0, 1, 1, 1),
                                      's3': (1, -1, 1, 1),
                                      's4': (1, 1, -1, 1)}
        self.bc_qp_coords_dict = {name: insert(bc_qp_coords, item[0], item[1], axis=1) for name, item in
                                  self.bc_surface_coord_dict.items()}
        self.diagram = IsoElementDiagram.quad4

    def set_quad8(self) -> None:
        self.dimension = 2
        self.topological_dimension = 2
        self.nodes_number = 8
        self.order = 3
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_quad8

        self.bc_surface_number = 4
        bc_quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension - 1)
        bc_qp_coords, self.bc_qp_weights = bc_quadrature.get_quadrature_coords_and_weights()
        self.bc_surface_nodes_dict = {'s1': (3, 0, 7),
                                      's2': (1, 2, 5),
                                      's3': (0, 1, 4),
                                      's4': (2, 3, 6)}
        self.bc_surface_coord_dict = {'s1': (0, -1, -1, 1),
                                      's2': (0, 1, 1, 1),
                                      's3': (1, -1, 1, 1),
                                      's4': (1, 1, -1, 1)}
        self.bc_qp_coords_dict = {name: insert(bc_qp_coords, item[0], item[1], axis=1) for name, item in
                                  self.bc_surface_coord_dict.items()}
        self.diagram = IsoElementDiagram.quad8

    def set_tria3(self) -> None:
        self.dimension = 2
        self.topological_dimension = 2
        self.nodes_number = 3
        self.order = 1
        quadrature = TriangleQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_tria3
        self.diagram = IsoElementDiagram.tria3

    def set_tria6(self) -> None:
        self.dimension = 2
        self.topological_dimension = 2
        self.nodes_number = 6
        self.order = 2
        quadrature = TriangleQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_tria6
        self.diagram = IsoElementDiagram.tria6

    def set_tetra4(self) -> None:
        tetra4 = r"""
        3
        * **
        *   * *
        *     *  *
        *       *   2
        *        **  *
        x2    *     * *
        |  x1         **
        0--x0 * * * * * 1"""

        self.dimension = 3
        self.topological_dimension = 3
        self.nodes_number = 4
        self.order = 1
        quadrature = TetrahedronQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_tetra4

        self.bc_surface_number = 4
        bc_quadrature = TriangleQuadrature(order=self.order, dimension=self.dimension - 1)
        bc_qp_coords, self.bc_qp_weights = bc_quadrature.get_quadrature_coords_and_weights()
        self.bc_surface_nodes_dict = {'s1': (0, 2, 3),
                                      's2': (0, 3, 1),
                                      's3': (0, 1, 2),
                                      's4': (1, 2, 3)}
        self.bc_surface_coord_dict = {'s1': (0, 0, -1, 1),
                                      's2': (1, 0, 1, 1),
                                      's3': (2, 0, -1, 1),
                                      's4': (0, 0, 1, sqrt(3))}
        self.bc_qp_coords_dict = {name: insert(bc_qp_coords, item[0], item[1], axis=1) for name, item in
                                  self.bc_surface_coord_dict.items()}
        s4_qp_coords = self.bc_qp_coords_dict['s4']
        s4_qp_coords[:, 0] = 1 - s4_qp_coords[:, 1] - s4_qp_coords[:, 2]
        self.diagram = IsoElementDiagram.tetra4

    def set_hex8(self) -> None:
        self.dimension = 3
        self.topological_dimension = 3
        self.nodes_number = 8
        self.order = 2
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_hex8

        self.bc_surface_number = 6
        bc_quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension - 1)
        bc_qp_coords, self.bc_qp_weights = bc_quadrature.get_quadrature_coords_and_weights()
        self.bc_surface_nodes_dict = {'s1': (0, 3, 7, 4),
                                      's2': (1, 2, 6, 5),
                                      's3': (0, 1, 5, 4),
                                      's4': (3, 2, 6, 7),
                                      's5': (0, 1, 2, 3),
                                      's6': (4, 5, 6, 7)}
        self.bc_surface_coord_dict = {'s1': (0, -1, -1, 1),
                                      's2': (0, 1, 1, 1),
                                      's3': (1, -1, 1, 1),
                                      's4': (1, 1, -1, 1),
                                      's5': (2, -1, -1, 1),
                                      's6': (2, 1, 1, 1)}
        self.bc_qp_coords_dict = {name: insert(bc_qp_coords, item[0], item[1], axis=1) for name, item in
                                  self.bc_surface_coord_dict.items()}
        self.diagram = IsoElementDiagram.hex8

    def set_hex20(self) -> None:
        self.dimension = 3
        self.topological_dimension = 3
        self.nodes_number = 20
        self.order = 3
        quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension)
        self.qp_coords, self.qp_weights = quadrature.get_quadrature_coords_and_weights()
        self.shape_function = get_shape_hex20

        self.bc_surface_number = 6
        bc_quadrature = GaussLegendreQuadrature(order=self.order, dimension=self.dimension - 1)
        bc_qp_coords, self.bc_qp_weights = bc_quadrature.get_quadrature_coords_and_weights()
        self.bc_surface_nodes_dict = {'s1': (0, 11, 3, 19, 7, 15, 4, 16),
                                      's2': (1, 9, 2, 18, 6, 13, 5, 17),
                                      's3': (0, 8, 1, 17, 5, 12, 4, 16),
                                      's4': (3, 10, 2, 18, 6, 14, 7, 19),
                                      's5': (0, 8, 1, 9, 2, 10, 3, 11),
                                      's6': (4, 12, 5, 13, 6, 14, 7, 15)}
        self.bc_surface_coord_dict = {'s1': (0, -1, -1, 1),
                                      's2': (0, 1, 1, 1),
                                      's3': (1, -1, 1, 1),
                                      's4': (1, 1, -1, 1),
                                      's5': (2, -1, -1, 1),
                                      's6': (2, 1, 1, 1)}
        self.bc_qp_coords_dict = {name: insert(bc_qp_coords, item[0], item[1], axis=1) for name, item in
                                  self.bc_surface_coord_dict.items()}
        self.diagram = IsoElementDiagram.hex20


iso_element_shape_dict: dict[str, IsoElementShape] = {
    'line2': IsoElementShape('line2'),
    'line3': IsoElementShape('line3'),
    'tria3': IsoElementShape('tria3'),
    'tria6': IsoElementShape('tria6'),
    'quad4': IsoElementShape('quad4'),
    'quad8': IsoElementShape('quad8'),
    'tetra4': IsoElementShape('tetra4'),
    'hex8': IsoElementShape('hex8'),
    'hex20': IsoElementShape('hex20')
}

if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(IsoElementShape.__slots_dict__)

    # iso_element_shape_dict['line2'].show()
    # iso_element_shape_dict['line3'].show()
    # iso_element_shape_dict['tria3'].show()
    # iso_element_shape_dict['tria6'].show()
    # iso_element_shape_dict['quad4'].show()
    # iso_element_shape_dict['quad8'].show()
    # iso_element_shape_dict['tetra4'].show()
    iso_element_shape_dict['hex8'].show()
    # iso_element_shape_dict['hex20'].show()



# -*- coding: utf-8 -*-
"""

"""
from typing import Dict, List

from pyfem.utils.colors import error_style
from pyfem.utils.visualization import object_slots_to_string


class BaseIO:
    """
    属性配置基类。

    当 self.is_read_only = True 时：BaseIO子类的所有属性在第一次被赋予非None值后变为只读状态，不能被修改或删除。

    """
    __slots__: List = []

    is_read_only: bool = True

    def __init__(self) -> None:
        pass

    def __setattr__(self, key, value):
        if self.is_read_only:
            if key not in self.__slots__:
                error_msg = f'\'{key}\' is not an allowable attribute keyword of \'{type(self).__name__}\''
                raise AttributeError(error_style(error_msg))
            elif hasattr(self, key) and self.__getattribute__(key) is not None:
                error_msg = f'attribute \'{type(self).__name__}.{key}\' is READ ONLY'
                raise PermissionError(error_style(error_msg))
            else:
                super().__setattr__(key, value)
        else:
            super().__setattr__(key, value)

    def __delattr__(self, key):
        if self.is_read_only:
            error_msg = f'attribute \'{type(self).__name__}.{key}\' is READ ONLY'
            raise PermissionError(error_style(error_msg))
        else:
            super().__delattr__(key)

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string(self, level)

    def show(self) -> None:
        print(self.to_string())

    def set_io_values(self, io_dict: Dict) -> None:
        for key, item in io_dict.items():
            self.__setattr__(key, item)


if __name__ == "__main__":
    io = BaseIO()
    io.show()



# -*- coding: utf-8 -*-
"""

"""
from numpy import ndarray, empty

from pyfem.assembly.Assembly import Assembly
from pyfem.io.Solver import Solver
from pyfem.utils.visualization import object_slots_to_string_ndarray


class BaseSolver:
    """
    求解器基类。

    :ivar assembly: 装配体对象
    :vartype assembly: Assembly

    :ivar solver: 求解器属性
    :vartype solver: Solver

    :ivar dof_solution: 求解得到自由度的值
    :vartype dof_solution: ndarray
    """

    __slots_dict__: dict = {
        'assembly': ('Assembly', '装配体对象'),
        'solver': ('Solver', '求解器属性'),
        'dof_solution': ('ndarray', '求解得到自由度的值')
    }

    __slots__: list = [slot for slot in __slots_dict__.keys()]

    def __init__(self) -> None:
        self.assembly: Assembly = None  # type: ignore
        self.solver: Solver = None  # type: ignore
        self.dof_solution: ndarray = empty(0)

    def to_string(self, level: int = 1) -> str:
        return object_slots_to_string_ndarray(self, level)

    def show(self) -> None:
        print(self.to_string())

    def run(self) -> int:
        return -1

    def solve(self) -> int:
        return -1


if __name__ == "__main__":
    from pyfem.utils.visualization import print_slots_dict

    print_slots_dict(BaseSolver.__slots_dict__)

    solver = BaseSolver()
    solver.show()



import setuptools
from src.pyfem import __version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyfem",
    version=__version__,
    author="Jingyu Sun",
    author_email="sun.jingyu@outlook.com",
    description="Python Finite Element Method",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sunwhale/pyfem",
    project_urls={
        "Bug Tracker": "https://github.com/sunwhale/pyfem/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    entry_points={
        'console_scripts': [
            'pyfem=pyfem.__main__:main'
        ]
    },
    install_requires=[
        'numpy',
        'scipy',
        'meshio',
        'tomli',
        'colorlog',
        'sympy'
    ],
)